import os

import lightning.pytorch as pl
import torch
from lightning.pytorch.callbacks import LearningRateMonitor, ModelCheckpoint
from models import SimCLR
from torch.utils import data

NUM_WORKERS = os.cpu_count()
CHECKPOINT_PATH = os.environ.get("PATH_CHECKPOINT", "saved_models/ContrastiveLearning/")
device = torch.device("cuda:0") if torch.cuda.is_available() else torch.device("cpu")


def train_simclr(batch_size, max_epochs=500, **kwargs):
    trainer = pl.Trainer(
        default_root_dir=os.path.join(CHECKPOINT_PATH, "SimCLR"),
        gpus=1 if str(device) == "cuda:0" else 0,
        max_epochs=max_epochs,
        callbacks=[
            ModelCheckpoint(save_weights_only=True, mode="max", monitor="val_acc_top5"),
            LearningRateMonitor("epoch"),
        ],
        progress_bar_refresh_rate=1,
    )
    trainer.logger._default_hp_metric = (
        None  # Optional logging argument that we don't need
    )

    # Check whether pretrained model exists. If yes, load it and skip training
    pretrained_filename = os.path.join(CHECKPOINT_PATH, "SimCLR.ckpt")
    if os.path.isfile(pretrained_filename):
        print(f"Found pretrained model at {pretrained_filename}, loading...")
        # Automatically loads the model with the saved hyperparameters
        model = SimCLR.load_from_checkpoint(pretrained_filename)
    else:
        train_loader = data.DataLoader(
            unlabeled_data,
            batch_size=batch_size,
            shuffle=True,
            drop_last=True,
            pin_memory=True,
            num_workers=NUM_WORKERS,
        )
        val_loader = data.DataLoader(
            train_data_contrast,
            batch_size=batch_size,
            shuffle=False,
            drop_last=False,
            pin_memory=True,
            num_workers=NUM_WORKERS,
        )
        pl.seed_everything(42)  # To be reproducable
        model = SimCLR(max_epochs=max_epochs, **kwargs)
        trainer.fit(model, train_loader, val_loader)
        # Load best checkpoint after training
        model = SimCLR.load_from_checkpoint(trainer.checkpoint_callback.best_model_path)

    return model



class ContrastiveTransformations:
    def __init__(self, base_transforms, n_views=2):
        self.base_transforms = base_transforms
        self.n_views = n_views

    def __call__(self, x):
        return [self.base_transforms(x) for i in range(self.n_views)]



import argparse
import pathlib

from dpat.convert import AvailableImageFormats, batch_convert


class BulkConvertArguments(argparse.Namespace):
    input_dir: str
    output_dir: str
    output_ext: AvailableImageFormats
    trust: bool
    skip_existing: bool
    num_workers: int
    chunks: int


def bulk_convert(args: BulkConvertArguments):
    ROOT_DIR = args.input_dir
    OUTPUT_DIR = args.output_dir
    OUTPUT_EXT = args.output_ext
    TRUST_SOURCE = args.trust
    SKIP_EXISTING = args.skip_existing
    NUM_WORKERS = args.num_workers
    CHUNKS = args.chunks

    paths = []
    kwargs_per_path = []
    output_dirs = []
    for scan_program in ["200slow", "300slow", "300fast"]:
        if scan_program == "200slow":
            resolution_unit: int = 3
            x_resolution: float = 5e4
            y_resolution: float = 5e4
        elif scan_program == "300slow":
            resolution_unit: int = 3
            x_resolution: float = 4e4
            y_resolution: float = 4e4
        elif scan_program == "300fast":
            resolution_unit: int = 3
            x_resolution: float = 1e4
            y_resolution: float = 1e4

        kwargs = dict(
            resolution_unit=resolution_unit,
            x_resolution=x_resolution,
            y_resolution=y_resolution,
        )
        add_paths = list(pathlib.Path(ROOT_DIR).glob(f"**/*{scan_program}*.bmp"))
        paths += add_paths
        output_dirs += [
            pathlib.Path(OUTPUT_DIR) / path.relative_to(ROOT_DIR).parent
            for path in add_paths
        ]
        kwargs_per_path += [kwargs] * len(add_paths)

    batch_convert(
        input_paths=paths,
        output_dirs=output_dirs,
        output_ext=OUTPUT_EXT,
        kwargs_per_path=kwargs_per_path,
        trust_source=TRUST_SOURCE,
        skip_existing=SKIP_EXISTING,
        num_workers=NUM_WORKERS,
        chunks=CHUNKS,
    )


def register_parser(parser: argparse._SubParsersAction):
    """Register convert commands to a root parser."""

    # Convert HHG images from any image to another format.
    convert_parser: argparse.ArgumentParser = parser.add_parser(
        "convert", help="Convert HHG images from any image to another format."
    )
    convert_subparsers = convert_parser.add_subparsers(help="Convert subparser.")
    convert_subparsers.required = True
    convert_subparsers.dest = "subcommand"

    bulk_convert_parser = convert_subparsers.add_parser(
        "bulk", help="Convert images in bulk."
    )
    bulk_convert_parser.add_argument(
        "--input-dir",
        "-i",
        type=str,
        required=True,
        help="Input directory where to find the images to be converted.",
        default=".",
    )
    bulk_convert_parser.add_argument(
        "--output-dir",
        "-o",
        type=str,
        help="Output directory where place converted files.",
        default="./converted",
    )
    bulk_convert_parser.add_argument(
        "--output-ext",
        "-e",
        type=str,
        choices=AvailableImageFormats.__members__,
        required=True,
        help="Extension to convert to.",
    )
    bulk_convert_parser.add_argument(
        "--num-workers",
        type=int,
        help="Number of workers that convert the images in parallel.",
        default=4,
    )
    bulk_convert_parser.add_argument(
        "--chunks",
        type=int,
        help="Number of chunks distributed to every worker.",
        default=30,
    )
    bulk_convert_parser.add_argument(
        "--trust",
        action=argparse.BooleanOptionalAction,
        help="Trust the source of the images.",
    )
    bulk_convert_parser.add_argument(
        "--skip-existing",
        action=argparse.BooleanOptionalAction,
        help="Skip existing output files.",
    )
    bulk_convert_parser.set_defaults(subcommand=bulk_convert)



import os
import urllib.request
from copy import deepcopy
from urllib.error import HTTPError

import lightning.pytorch as pl
import matplotlib
import matplotlib.pyplot as plt
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data as data
import torchvision
from torchvision import transforms
from tqdm import tqdm
from train import train_simclr
from transformations import ContrastiveTransformations


def main():
    # Path to the folder where the datasets are/should be downloaded (e.g. CIFAR10)
    DATASET_PATH = os.environ.get("PATH_DATASETS", "data/")
    # Path to the folder where the pretrained models are saved
    CHECKPOINT_PATH = os.environ.get(
        "PATH_CHECKPOINT", "saved_models/ContrastiveLearning/"
    )
    # In this notebook, we use data loaders with heavier computational processing. It is recommended to use as many
    # workers as possible in a data loader, which corresponds to the number of CPU cores
    NUM_WORKERS = os.cpu_count()

    # Setting the seed
    pl.seed_everything(42)

    # Ensure that all operations are deterministic on GPU (if used) for reproducibility
    torch.backends.cudnn.determinstic = False
    torch.backends.cudnn.benchmark = False

    device = (
        torch.device("cuda:0") if torch.cuda.is_available() else torch.device("cpu")
    )
    print("Device:", device)
    print("Number of workers:", NUM_WORKERS)

    # Create checkpoint path if it doesn't exist yet
    os.makedirs(CHECKPOINT_PATH, exist_ok=True)

    # TODO: CREATE TILE DATASETS

    simclr_model = train_simclr(
        batch_size=256,
        hidden_dim=128,
        lr=5e-4,
        temperature=0.07,
        weight_decay=1e-4,
        max_epochs=500,
    )


if __name__ == "__main__":
    main()



"""DPAT Command-line interface. This is the file which builds the main parser."""
import argparse


def main():
    """
    Console script for dlup.
    """
    # From https://stackoverflow.com/questions/17073688/how-to-use-argparse-subparsers-correctly
    root_parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    root_subparsers = root_parser.add_subparsers(help="Possible DPat CLI utils to run.")
    root_subparsers.required = True
    root_subparsers.dest = "subcommand"

    # Prevent circular import
    from dpat.cli.convert import register_parser as register_convert_subcommand
    from dpat.cli.create_splits import register_parser as register_splits_subcommand

    # Whole slide images related commands.
    register_convert_subcommand(root_subparsers)
    register_splits_subcommand(root_subparsers)

    args = root_parser.parse_args()
    args.subcommand(args)


if __name__ == "__main__":
    main()



import lightning.pytorch as pl
import torch
import torchvision
from torch import nn, optim
from torch.functional import F


class SimCLR(pl.LightningModule):
    def __init__(self, hidden_dim, lr, temperature, weight_decay, max_epochs=500):
        super().__init__()
        self.save_hyperparameters()
        assert (
            self.hparams["temperature"] > 0.0
        ), "The temperature must be a positive float!"
        # Base model f(.)
        # TODO: SWAP OUT RESNET FOR SHUFFLENETV2
        self.convnet = torchvision.models.resnet18(
            pretrained=False, num_classes=4 * hidden_dim
        )  # num_classes is the output size of the last linear layer
        # The MLP for g(.) consists of Linear->ReLU->Linear
        self.convnet.fc = nn.Sequential(
            self.convnet.fc,  # Linear(ResNet output, 4*hidden_dim)
            nn.ReLU(inplace=True),
            nn.Linear(4 * hidden_dim, hidden_dim),
        )

        self.loss_fn = nn.CrossEntropyLoss()

    def configure_optimizers(self):
        optimizer = optim.AdamW(
            self.parameters(),
            lr=self.hparams["lr"],
            weight_decay=self.hparams["weight_decay"],
        )
        lr_scheduler = optim.lr_scheduler.CosineAnnealingLR(
            optimizer, T_max=self.hparams["max_epochs"], eta_min=self.hparams["lr"] / 50
        )
        return [optimizer], [lr_scheduler]

    def training_step(self, batch, batch_idx):
        loss = self.loss_fn(batch)
        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch, batch_idx):
        loss = self.loss_fn(batch)
        self.log("train_loss", loss)



import argparse

from dpat.splits.create_splits import create_splits


class CreateSplitsArguments(argparse.Namespace):
    input_dir: str
    output_dir: str
    path_to_labels_file: str
    dataset_name: str
    filter_diagnosis: list[str]
    overwrite: bool
    include: list[str]
    exclude: list[str]


def cli_create_splits(args: CreateSplitsArguments) -> None:
    create_splits(
        image_dir=args.input_dir,
        path_to_labels_file=args.path_to_labels_file,
        dataset_name=args.dataset_name,
        save_to_dir=args.output_dir,
        overwrite=args.overwrite,
        include_pattern=args.include,
        exclude_pattern=args.exclude,
        filter_diagnosis=args.filter_diagnosis,
    )


def register_parser(parser: argparse._SubParsersAction):
    """Register splits commands to a root parser."""

    # Create splits
    splits_parser: argparse.ArgumentParser = parser.add_parser(
        "splits", help="Create data splits."
    )
    splits_subparsers = splits_parser.add_subparsers(help="Splits subparser.")
    splits_subparsers.required = True
    splits_subparsers.dest = "subcommand"

    create_splits_parser = splits_subparsers.add_parser(
        "create", help="Create data splits."
    )
    create_splits_parser.add_argument(
        "--input-dir",
        "-i",
        type=str,
        required=True,
        help="Input directory where to find the images.",
        default=".",
    )
    create_splits_parser.add_argument(
        "--output-dir",
        "-o",
        type=str,
        required=True,
        help="Name of dataset.",
        default="splits",
    )
    create_splits_parser.add_argument(
        "--labels",
        "-l",
        type=str,
        required=True,
        help="Path to labels file.",
        default="labels.csv",
        dest="path_to_labels_file",
    )
    create_splits_parser.add_argument(
        "--name",
        "-n",
        type=str,
        required=True,
        help="Name of dataset.",
        dest="dataset_name",
    )
    create_splits_parser.add_argument(
        "--filter",
        "-f",
        type=str,
        help="Filter a diagnosis. For multiple diagnoses, use `-f 1 -f 2`.",
        dest="filter_diagnosis",
        action="append",
    )
    create_splits_parser.add_argument(
        "--overwrite",
        action=argparse.BooleanOptionalAction,
        help="Overwrite folds in output dir, if available.",
    )
    create_splits_parser.add_argument(
        "--include",
        "-y",
        type=str,
        help="Glob pattern to include files from `input-dir`",
        default=["*.*"],
        dest="include",
        action="append",
    )
    create_splits_parser.add_argument(
        "--exclude",
        "-x",
        type=str,
        help="Glob pattern to exclude files from `input-dir`",
        default=[""],
        dest="exclude",
        action="append",
    )
    create_splits_parser.set_defaults(subcommand=cli_create_splits)



import os
import platform

import yaml

with open("config.yml", "r") as f:
    config = yaml.safe_load(f)

# vips must be installed separately for Windows.
# vips already includes OpenSlide.
# Provide the path to vips\bin in project.ini.
# https://github.com/libvips/pyvips
if platform.system() == "Windows":
    PYVIPS_PATH = config["PATHS"]["vips"]
    os.environ["PATH"] = PYVIPS_PATH + ";" + os.environ["PATH"]
    try:
        import pyvips  # isort:skip
        import openslide
    except OSError:
        raise ImportError(
            "Make sure to install vips and set PATHS.vips in project.ini."
        )

import dlup



import pathlib
import signal
from enum import Enum
from functools import partial
from multiprocessing import Pool
from typing import Literal, Optional, TypedDict

from PIL import Image
from tqdm import tqdm


class AvailableImageFormats(Enum):
    tiff = "tiff"
    tif = "tif"


class ToTIFFParams(TypedDict):
    """Allowed parameters for `img_to_tiff`.

    TIFF files [1] contain resolution information in the header.

    Attributes
    ----------
    resolution_unit : int
        Resolution unit, indicating no (1), inch (2) or cm (3) as unit.
    xresolution : float
        Number of pixels per resolution unit in X-direction.
    yresolution : float
        Number of pixels per resolution unit in Y-direction.

    References
    ----------
    [1] https://www.fileformat.info/format/tiff/corion.htm
    """

    resolution_unit: Literal[1, 2, 3]
    x_resolution: float
    y_resolution: float


class ToOtherParams(ToTIFFParams):
    pass


def img_to_tiff(
    input_path: pathlib.Path,
    output_dir: pathlib.Path,
    extension: AvailableImageFormats,
    trust_source: bool = False,
    **kwargs: ToTIFFParams,
) -> None:
    """Convert any image to TIFF.

    Parameters
    ----------
    input_path : pathlib.Path
        Path to the input image.
    output_dir : pathlib.Path
        Directory to output image in.
    extension : str
        Output file extension.
    trust_source : bool
        Trust source. If True, PIL.Image.DecompressionBombError is ignored.
    kwargs : `ToTIFFParams`
        Extra arguments to `PIL.Image.save`: refer to the `ToTIFFParams` documentation for a list of all possible arguments.
    """
    # assert not input_path.endswith([".tiff", ".tif"]), "Input image is already in TIFF format."

    # To ignore decompression bomb DOS attack error.
    if trust_source:
        Image.MAX_IMAGE_PIXELS = None

    output_fn = pathlib.Path(output_dir / (input_path.stem + f".{extension}"))
    output_dir.mkdir(parents=True, exist_ok=True)

    bmp_image = Image.open(input_path)
    bmp_image.save(output_fn, **kwargs)
    bmp_image.close()


def filter_existing(
    input_paths: list[pathlib.Path],
    output_dirs: list[pathlib.Path],
    extension: AvailableImageFormats,
    kwargs_per_path: Optional[list[ToOtherParams]] = None,
):
    filtered_input_paths = []
    filtered_output_dirs = []
    filtered_kwargs_per_path = []

    for input_path, output_dir, kwargs in zip(
        input_paths, output_dirs, kwargs_per_path
    ):
        output_fn = pathlib.Path(output_dir / (input_path.stem + f".{extension}"))
        if not output_fn.exists():
            filtered_input_paths.append(input_path)
            filtered_output_dirs.append(output_dir)
            filtered_kwargs_per_path.append(kwargs)

    skip_count = len(input_paths) - len(filtered_input_paths)
    print(
        f"Skipping {skip_count}/{len(input_paths)} images, as they were already converted to {extension}. "
        "Use '--skip-existing false' to overwrite existing images."
    )

    return filtered_input_paths, filtered_output_dirs, filtered_kwargs_per_path


def _wrapper(args, worker):
    input_path, output_dir, kwargs = args
    worker(input_path, output_dir, **kwargs)


def batch_convert(
    input_paths: list[pathlib.Path],
    output_dirs: list[pathlib.Path],
    output_ext: AvailableImageFormats,
    num_workers: int = 4,
    chunks: int = 30,
    kwargs_per_path: Optional[list[ToOtherParams]] = None,
    trust_source=False,
    skip_existing=True,
) -> None:
    """Convert all images found in the input directory.

    Parameters
    ----------
    input_paths : str
        Path to the input directory.
    output_dir : str
        Path to the output directory.
    output_ext : str
        Output extension
    trust_source : bool
        Trust source. If True, PIL.Image.DecompressionBombError is ignored.
    skip_existing : bool
        Skip existing output files.
    kwargs : `ToOtherParams`
        Extra arguments to `convert_func`.
    """

    if output_ext in ["tiff", "tif"]:
        convert_func = partial(
            img_to_tiff,
            extension=output_ext,
            skip_existing=skip_existing,
            trust_source=trust_source,
        )

    if skip_existing:
        input_paths, output_dirs, kwargs_per_path = filter_existing(
            input_paths,
            output_dirs,
            output_ext,
            kwargs_per_path,
        )

    if not input_paths:
        print("No images to convert.")
        return

    wrapper = partial(_wrapper, worker=convert_func)

    nmax = len(input_paths)
    chunksize = max(nmax // chunks, 1)
    with Pool(
        num_workers, initializer=signal.signal, initargs=(signal.SIGINT, signal.SIG_IGN)
    ) as pool:
        try:
            list(
                tqdm(
                    pool.imap_unordered(
                        wrapper,
                        zip(input_paths, output_dirs, kwargs_per_path),
                        chunksize,
                    ),
                    total=nmax,
                    desc="Converting images",
                )
            )
        except KeyboardInterrupt:
            print("Interrupted.")



from setuptools import find_packages, setup  # type: ignore

__version__ = "1.1.0"

# Get the long description from the README file
with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

setup(
    author="Siem de Jong",
    author_email="siem.dejong@hotmail.nl",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    python_requires=">=3.9",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    entry_points={
        "console_scripts": [
            "dpat=dpat.cli:main",
        ],
    },
    install_requires=[
        "dlup>=0.3",
        "scikit-learn>=0.24",
    ],
    license="GNU General Public License v3",
    keywords="dpat",
    name="dpat",
    packages=find_packages(include=["dpat", "dpat.*"]),
    url="https://github.com/siemdejong/hhg-dpat",
    version=__version__,
)



import pathlib

from dlup.data.dataset import ConcatDataset, TiledROIsSlideImageDataset


class PMCHHGTileDataset(ConcatDataset):
    """
    add documentation on how this dataset works
    Args:
        add docstrings for the parameters
    """

    def __init__(self, cfg, path, split, dataset_name, data_source):
        self.path = pathlib.Path(path)
        super().__init__([TiledROIsSlideImageDataset(fn) for fn in self.path.glob("*")])

    def num_samples(self):
        """
        Size of the dataset
        """
        return self.__len__()

    def __getitem__(self, idx: int):
        """
        implement how to load the data corresponding to idx element in the dataset
        from your data source
        """
        tile_dict = ConcatDataset.__getitem__(self, idx)

        return tile_dict["image"]



import lightning.pytorch as pl
from torch.utils.data import DataLoader


class PMCHHGDataModule(pl.LightningDataModule):
    def train_dataloader(self):
        return DataLoader(self.PMC_HHG_train, batch_size=64)

    def val_dataloader(self):
        return DataLoader(self.PMC_HHG_val, batch_size=64)

    def test_dataloader(self):
        return DataLoader(self.PMC_HHG_test, batch_size=64)



class DpatOutputDirectoryExistsError(Exception):
    def __init__(self, path, *args):
        super().__init__(args)
        self.path = path

    def __str__(self):
        return f"Output directory '{self.path}' already exists. Choose another or use --overwrite."
